#!/usr/bin/env python
import datetime
import numpy as np
import os.path
import sys
from os.path import basename, dirname

from libinsitu.common import *
from libinsitu.cdl import *
from libinsitu.handlers import HANDLERS, InSituHandler, listNetworks
from libinsitu.log import debug, info, warning, logger, LogContext, error
import argparse


DONE_SUFFIX = '.done'
ERR_SUFFIX = '.err'

BOUNDARY_TOLERANCE_PCT = 5

def check_boundaries(var, data) :
    """Check boundaries of a variable"""
    if not hasattr(var, VALID_MIN_ATTR) :
        return

    debug("Checking boundaries for %s" % var)

    minv = parse_value(getattr(var, VALID_MIN_ATTR))
    maxv = parse_value(getattr(var, VALID_MAX_ATTR))

    range = maxv - minv

    # Add tolerance
    minv = minv - range * BOUNDARY_TOLERANCE_PCT / 100
    maxv = maxv + range * BOUNDARY_TOLERANCE_PCT / 100

    def warn_boundary(idx, sign, bound) :
        if np.any(idx):
            warning("Boundary check : %d items of %s are %s %f: [%f:%f]",
                    np.sum(idx),
                    var.name,
                    sign,
                    bound,
                    np.min(data[idx]),
                    np.max(data[idx]))

    warn_boundary(data < minv, "<", minv)
    warn_boundary(data > maxv, ">", maxv)


def list_files(in_files, handler) :

    # Gather and sort files with pattern
    files = []
    for file_or_dir in in_files:
        if os.path.isdir(file_or_dir):
            files += handler.list_files(file_or_dir)
        else:
            files.append(file_or_dir)

    # Sort input files
    in_files = handler.sort_files(files)

    if len(in_files) == 0:
        warning("No input file found")
        sys.exit(0)

    return in_files


def process_network(network, station_id, out_filename, args) :

    # Get properties for this station
    properties = getProperties(network, station_id)

    now = datetime.now().isoformat()

    properties["UpdateTime"] = now
    properties["CreationTime"] = now

    handler : InSituHandler = HANDLERS[network](properties)

    in_files = list_files(args.in_files, handler)

    new = not os.path.exists(out_filename)
    mode = "w" if new else "a"
    ncfile = Dataset(out_filename, mode=mode)

    if  new :
        # Nc File does not exist ==> create it
        info("File '%s' was not there. Initializing it.", out_filename)
        init_nc(ncfile, properties, [])

    # Loop on input files
    for infile in in_files :

        # Incremental mode : check status files
        status_folder = args.status_folder or dirname(infile)
        status_file = os.path.join(status_folder, basename(infile) + DONE_SUFFIX)
        err_file = os.path.join(status_folder, basename(infile) + ERR_SUFFIX)

        # Incremental mode : if output was already there, don't proess input files having a more recent .done file
        if not new and args.incremental and os.path.exists(status_file) and older_than(infile, status_file):
            info("File %s is older than status file %s : Skipping", infile, status_file)
            continue

        with LogContext(file=os.path.basename(infile)):

            # Safe execution : do not stop on error
            try:
                process_chunck(handler, infile, ncfile, args, properties)

                # Incremental mode : touch status file
                if args.incremental:
                    touch(status_file)

                    # err file was present : delete it
                    if os.path.exists(err_file) :
                        os.remove(err_file)

            # Don't intercept Ctrl-C : cancel the whole process
            except KeyboardInterrupt as e :
                raise e

            except Exception as e :

                # Write .err file
                if args.incremental:
                    touch(os.path.join(status_folder, basename(infile) + ERR_SUFFIX))

                # Do not fail : just log and process the next file
                logger.exception(e)



def idx2slice(idx) :
    """If indices are regular, transform it into a slice : much faster for writing"""

    if len(idx) < 2 :
        return idx

    step = idx[1] - idx[0]

    if np.all(np.diff(idx) == step) :
        debug("Idx was slice")
        return slice(idx[0], idx[-1] + step, step)

    return idx

def check_and_assign(ncfile, data, times_idx, size_before, args) :

    # Check once for all if new chunk overlaps
    overlapping_mask = times_idx < size_before
    overlapping_indices = times_idx[overlapping_mask]

    for varname in data.columns:

        var = ncfile.variables[varname]
        new_values = data[[varname]].values.flatten()

        check_boundaries(var, new_values)

        # Overcoming this bug ; https://github.com/scipy/scipy/issues/6097
        if np.issubdtype(var.dtype, np.integer) :
            fill_value = getattr(var, FILL_VALUE_ATTR, DEFAULT_FILL_VALUE)
            new_values[np.isnan(new_values)] = fill_value

        if not np.any(overlapping_mask) or not args.check:
            # No overlap with previous data ? no need for check
            var[idx2slice(times_idx)] = new_values

        else :
            debug("Possible overlap, checking ...")

            if np.all(np.isnan(var[overlapping_indices])) :

                # All nans ? don't check further"
                write_mask = np.ones(new_values.shape, dtype=bool)

            else:
                # We won't override existing values with nans
                nan_mask = np.isnan(new_values)

                nna_overlapping_mask = overlapping_mask & ~nan_mask
                overlapping_idx = times_idx[nna_overlapping_mask]
                overlapping_values = new_values[nna_overlapping_mask]
                overlapped_values = var[overlapping_idx]

                conflicting_indices = ~np.isnan(overlapped_values) & ~np.isclose(overlapping_values, overlapped_values)

                if np.any(conflicting_indices) :

                    conflicting_overlapped =  overlapped_values[conflicting_indices]
                    conflicting_overlapping = overlapping_values[conflicting_indices]

                    warning("%d conflicting values for %s. Overriding [%f:%f] -> [%f:%f]",
                            np.sum(conflicting_indices),
                            varname,
                            np.nanmin(conflicting_overlapped), np.nanmin(conflicting_overlapped),
                            np.nanmin(conflicting_overlapping), np.nanmin(conflicting_overlapping))

                # Write NaN if they do not overlap with existing data (this extends the dimension)
                write_mask = nan_mask | ~overlapping_mask

            var[idx2slice(times_idx[write_mask])] = new_values[write_mask]



def process_chunck(handler, infile, ncfile, args, properties):

    info("processing chunk : %s", infile)

    # Read data
    data = handler.read_chunk(infile)

    if data is None or len(data) == 0 :
        warning("Chunk is empty")
        return

    # Time resolution, in seconds
    resolution_s = getTimeResolution(ncfile)

    # Drop duplicates
    data = data[~data.index.duplicated(keep="last")]

    # Reshape : regular time is faster to write in NetCDF (as slice)
    data = data.asfreq("%dS" % resolution_s)

    # Transform time to seconds since start date and time idx
    chunk_dates = data.index.values

    times_sec = datetime64_to_sec(ncfile, chunk_dates)

    columns = list(data.columns)

    # Create vars if not present yet
    missing_vars = list(col for col in columns if not col in ncfile.variables)
    if len(missing_vars) > 0:
        info("Adding missing vars : %s", missing_vars)
        init_nc(ncfile, properties, missing_vars)

    # Ensure all timestamps fall into resolution
    exact_idx = (times_sec % resolution_s) == 0
    nb_not_exact_times = len(times_sec) - np.sum(exact_idx)
    if nb_not_exact_times > 0:

        # Remove lines with wrong times
        data = data[exact_idx]
        times_sec = times_sec[exact_idx]

        warning("%d rows had timings not fitting the time resolution : skipping them" % nb_not_exact_times)


    # Seconds to time index, as per start date and resolution
    time_idx = seconds_to_idx(ncfile, times_sec)

    chunk_start = min(chunk_dates)
    chunk_end = max(chunk_dates)
    chunk_end_int = datetime64_to_sec(ncfile, chunk_end)

    info("chunck range: %s to %s. samples:%d", time2str(chunk_start), time2str(chunk_end), len(data.index))

    # Error if chunk starts before start time
    if sum(time_idx < 0) > 0 :
        raise Exception("Chunk start (%s) is before output start time. Skipping" % (time2str(chunk_start)))

    # Warning if resolution seems different
    # Error if scrictREsolution is set
    if len(times_sec) >= 2:

        periods = get_periods(times_sec)
        if len(periods) >= 1 :
            actual_resolution, count = periods[0]

            if actual_resolution != resolution_s:
                warning("Resolution of input chunk (%d sec) differs from resolution of output (%d sec)" % (actual_resolution, resolution_s))
                if args.strict_resolution :
                    warning("Strict resolution requested : skipping")
                    return

    # Fill time variable with proper values
    size_before = len(ncfile.variables[TIME_VAR]) # Remember the size of TIME before it is extended

    # Fill Time variable
    if len(ncfile.variables[TIME_VAR]) == 0 :
        next_time_sec =  datetime64_to_sec(ncfile, start_date64(ncfile))
    else:
        next_time_sec = ncfile.variables[TIME_VAR][-1] + resolution_s

    end_time_sec = chunk_end_int + resolution_s
    new_times_sec = np.arange(next_time_sec, end_time_sec, resolution_s)
    next_time_idx = seconds_to_idx(ncfile, next_time_sec)
    end_time_idx = seconds_to_idx(ncfile, end_time_sec)
    ncfile.variables[TIME_VAR][next_time_idx: end_time_idx] = new_times_sec

    # Store data values
    check_and_assign(ncfile, data, time_idx, size_before, args)

    info("Chunk processed successfully")


def sort_files(files) :
    """ Sort filenames named like xxxMMYY*"""
    def yearmonth(path):
        file = basename(path)
        year = file[5:7]
        month = file[3:5]
        if year.isnumeric() :
            year_num = int(year)
            year_num += 1900 if year_num > 70 else 2000
            year = str(year_num)
        res =  year + '-' + month + '-' + file[7:]
        return res

    return sorted(files, key=yearmonth)

def dir_path(path):
    if os.path.isdir(path):
        return path
    else:
        raise argparse.ArgumentTypeError(f"{path} is not a valid folder")


def parser() :
    parser = argparse.ArgumentParser(description='Transforms In-Situ data into NetCDF files')
    parser.add_argument('out', metavar='<out.nc>', type=str, help='Output file')
    parser.add_argument('in_files', metavar='<file|dir>', nargs='+', help='Input files or folders')
    parser.add_argument('--network', '-n', help='Network name', required=True, choices=listNetworks())
    parser.add_argument('--station-id', '-s', metavar='<SID>', help='Station ID', required=True)
    parser.add_argument('--incremental', '-i', default=False, action='store_true',
                        help="Incremental mode, skipping input files having a '.done' status files")
    parser.add_argument('--strict-resolution', '-sr', default=False, action='store_true',
                        help="Skip chunks having a different resulution")
    parser.add_argument('--check', '-c', default=False, action='store_true', help="Check potential override of data")
    parser.add_argument('--status-folder', '-f', metavar='<folder>', type=dir_path,
                        help='Separate folder for .done/.err files')
    return parser

def main():

    args = parser().parse_args()

    network = args.network
    station_id  = args.station_id

    with LogContext(network=network, station_id=station_id):
        process_network(network, station_id, args.out, args)

if __name__ == '__main__':
    main()


from libinsitu.common import netcdf_to_dataframe, get_periods
from libinsitu.log import *
import numpy as np

def check_time(filename) :

    df = netcdf_to_dataframe(filename)
    station_id = df.attrs["StationInfo_Abbreviation"]
    network = df.attrs["source"]
    if " " in network :
        network = network.split(" ")[0]

    with LogContext(station_id=station_id, network=network, file=filename) :

        from_date = min(df.index)
        to_date = max(df.index)
        nb_samples = len(df.index)

        time_s = df.index.values.astype(np.int64) // 1000000000

        periods = get_periods(time_s)

        if len(periods) > 1 :
            warning("Found several periods periods=%s", periods)
        else:
            periods = list(periods.keys())[0]
        info("from:%s, to:%s, %d samples, period:%s", from_date, to_date, nb_samples, periods)

def main() :
    """Perform various checks on a InSitu NetCDF file"""
    for file in sys.argv[1:] :
        with IgnoreAndLogExceptions() :
            check_time(filename=file)

if __name__ == '__main__':
    main()



import argparse

from pandas import DataFrame

from libinsitu import getProperties, getNetworksInfo, getStationInfo, getStationsInfo, VALID_COLS, df_to_csv, df_to_json
from libinsitu.handlers import listNetworks
import sys

def parser() :
    parser = argparse.ArgumentParser(description='Prints / export meta data about networks and stations')
    parser.add_argument('--format', '-f', metavar='<format>', help="Output format", choices=["txt", "csv", "json"], default="txt")
    parser.add_argument('--no-header', '-nh', action="store_true", help="Disable printing of header for txt and csv output", default=False)
    parser.add_argument('--columns', '-c', metavar='<col1>,<col2>,...', help="Columns to show. All by default.")

    sub_parsers = parser.add_subparsers(dest="command", required=True)
    parser_networks = sub_parsers.add_parser("networks", help="Show networks meta-data")

    parser_stations = sub_parsers.add_parser("stations", help="Show networks meta-data")
    parser_stations.add_argument('--network', '-n', metavar='<network-id>', help="Only show stations for the selected network")
    return parser


def main() :


    args = parser().parse_args()

    if args.command == "networks" :
        data = getNetworksInfo().values()
    else :

        if args.network :
            data = [attrs for attrs in getStationsInfo(args.network).values()]
        else:
            data = []
            for network in listNetworks():
                data += [dict(network=network, **val) for val in getStationsInfo(network).values()]


    df = DataFrame(data, dtype=str).fillna("")

    if args.columns :
        columns = args.columns.split(",")
        df = df[columns]

    # Reorder columns
    if args.command == "stations" :
        df = df[[col for col in ["network"] + VALID_COLS if col in df.columns]]

    if args.format == "txt" :
        df.to_string(sys.stdout, index=False, header=not args.no_header)
        print()
    elif args.format == "csv" :
        df_to_csv(df, index=False, header=not args.no_header)
    else:
        df_to_json(df, orient="records", indent=2)




import os
from setuptools import setup, find_packages
import subprocess
from datetime import datetime
import pkg_resources
import pkgutil

# Utility function to read the README file.
# Used for the long_description.  It's nice, because now 1) we have a top level
# README file and 2) it's easier to type in the README file than to put a raw
# string in below ...
def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname))\
        .read().strip()

def run(args) :
    return subprocess.run(args, stdout=subprocess.PIPE).\
        stdout.decode('utf-8').splitlines()

# Get current git branch
branches = run(["git", "branch"])
curr_branch = next(line for line in branches if "*" in line)
curr_branch = curr_branch.replace(" ", "").replace("*", "")
version = read("VERSION")
name = "libinsitu"


if curr_branch == "dev" :

    name += "_dev"

    #commit = run(["git", "log"])[0].split()[1][0:8]

    start = datetime.strptime("2021-01-01", '%Y-%m-%d')
    now = datetime.now()

    min_diff = int((now-start).total_seconds() // 60)

    version += "." + str(min_diff) + "-dev"


extra_urls= []
with open("requirements.txt", "r") as f :

    def extract_extra_index(strs) :
        for line in strs :
            if line.startswith("--extra-index-url") :
                _, url, rest = line.split(" ")
                extra_urls.append(url)
                yield rest
            else:
                yield line

    requirements = [
            str(requirement).replace("==", ">=")
            for requirement
            in pkg_resources.parse_requirements(extract_extra_index(f))]



# List all cli modules
import libinsitu.cli
entry_points = []
for importer, modname, ispkg in pkgutil.iter_modules(libinsitu.cli.__path__):
    entry_points.append('ins-%s = libinsitu.cli.%s:main' % (modname, modname))
print("entry points :", entry_points)

packages = find_packages()

print("Packages : %s"%  str(packages))
print("Extra URLs : %s" % extra_urls)

setup(
    name = name,
    version = version,
    python_requires='>3.7',
    author = "OIE - Mines ParisTech",
    author_email = "raphael.jolivet@mines-paristech.fr",
    description = ("This library provides tools to transform solar irradiation data from various networks to uniform NetCDF files. "
                   "It also provides tools to request and manipulate those NetCDF files"),
    license = "BSD",
    keywords = "in-situ, solar, pv, irradiation, NetCDF, FAIR, meta-data",
    url = "https://git.sophia.mines-paristech.fr/oie/libinsitu",
    packages=packages,
    long_description=read('README.md'),
    long_description_content_type='text/markdown',
    dependency_links=extra_urls,
    include_package_data=True,
    classifiers=[],
    install_requires=requirements,
    entry_points={'console_scripts': entry_points}
)





# Handling fetch and parsing of Thredds server

import xml.etree.ElementTree as ET

from typing import Dict
from urllib.parse import urljoin, urlsplit
import pprint

from requests import HTTPError, Session

from libinsitu import parallel_map
from libinsitu.log import info

NS={
    "thredds" : "http://www.unidata.ucar.edu/namespaces/thredds/InvCatalog/v1.0",
    "xlink" : "http://www.w3.org/1999/xlink"}

XLINK = "{http://www.w3.org/1999/xlink}"


class AutoRepr :
    def __repr__(self):
        return pprint.pformat(self.__dict__)

class Dataset(AutoRepr) :
    def __init__(self, id, name, size) :
        self.id = id
        self.name = name
        self.size = size
        self.services = dict() # Dict of service type => URL

class Catalog(AutoRepr) :
    def __init__(self, id, name, description, url, link) :
        self.id = id
        self.name = name
        self.description = description
        self.url = url
        self.link = link
        self.authorized = None
        self.datasets : Dict[str, Dataset]= dict()
        self.catalogs : Dict[str, Catalog] = dict() # Dict of sub catalogs
        self.services = dict() # Dict of service type => URL

def parse_services(catalog, base):

    res = dict()
    for service in catalog.findall('.//thredds:service', NS):
        name = service.get("name")
        url = service.get("base")
        if url != "":
            res[name] = base + url
    return res

def parse_catalog(el, url, id=None, name=None) :
    """Parse catalog from 'Dataset' element of root catalog or datasetREf/metadata element """

    if id is None :
        name = el.attrib['name']
        id = el.attrib['ID']

    # find doc
    metadata = el.find('thredds:metadata', NS)  # Optionnal nested node "metadata"
    if metadata is None:
        metadata = el
    docs = metadata.findall('thredds:documentation', NS)
    description = None
    link = None
    for doc in docs:
        if doc.get("type") == "summary":
            description = doc.text
        else:
            link_ = doc.get('{%s}href' % NS["xlink"])
            if link_ is not None:
                link = link_
    return Catalog(id, name, description, url, link)

def extract_sub(url, subCatEl) :
    id = subCatEl.attrib[XLINK + "title"]
    href = subCatEl.attrib[XLINK + "href"]
    sub_url = urljoin(url, href)
    return (id, sub_url)

def fetch_catalog(url, session=None, recursive=True, parallel=True) :

    info("Fetching : %s" % url)

    if session is None :
        session = Session()

    base = base_url(url)
    xml = http_get(url, session)

    catalogEl = ET.fromstring(xml)
    datasetEl = catalogEl.find('thredds:dataset', NS)

    catalog = parse_catalog(datasetEl, url)
    catalog.services = parse_services(catalogEl, base)

    # Parse or fetch sub catalogs

    sub_elements = list(datasetEl.findall("thredds:catalogRef", NS))

    def fetch_rec(subCatEl) :
        id, sub_url = extract_sub(url, subCatEl)

        try:
            sub_catalog = fetch_catalog(sub_url, session, parallel=parallel)
            return (id, sub_catalog)
        except HTTPError as e:
            if e.response.status_code == 401:
                print("URL %s not authorized." % e.request.url)
            else:
                raise e

    if recursive :
        # Common loop for sequential of parallel fetch
        for id, sub_catalog in parallel_map(fetch_rec, sub_elements, parallel):
            catalog.catalogs[id] = sub_catalog
    else :
        for subCatEl in sub_elements :
            id, sub_url = extract_sub(url, subCatEl)
            catalog.catalogs[id] = parse_catalog(subCatEl, sub_url, id, id)

    # Datasets
    datasets = datasetEl.findall("thredds:dataset", NS)

    for dataset_el in datasets :

        uri = dataset_el.attrib["urlPath"]
        id = dataset_el.attrib["ID"]
        name = dataset_el.attrib["name"]
        size = float(dataset_el.find("thredds:dataSize", NS).text)

        dataset = Dataset(id, name, size)
        dataset.services = {key: base + uri for key, base in catalog.services.items()}

        catalog.datasets[dataset.name] = dataset

    return catalog

def base_url(url) :
    parts = urlsplit(url)
    return  "%s://%s" % (parts.scheme, parts.netloc)



def http_get(url, session) :

    #print("fetching : %s" % url)

    res = session.get(url)
    res.raise_for_status()

    res.encoding = res.apparent_encoding
    return res.text


import numpy as np
from matplotlib import pyplot as plt
from matplotlib.gridspec import GridSpec

from libinsitu import info
from libinsitu.qc.matplot.graphs import BaseMatplotlibGraphs, MC_CLEAR_COLOR


class MaplotLibGraphs(BaseMatplotlibGraphs):

    def main_layout(self) :
        """ Render main layout """

        info("QC: visual plot preparation")
        fig = plt.figure(figsize=(19.2, 9.93))

        # =====================================================================
        # First column : time series + heatmaps + ratios
        # =====================================================================

        # Draw grid
        grid = GridSpec(8 if self.cams_df is None else 9, 1)
        grid.update(
            left=0.035, right=0.32,
            bottom=0.03, top=0.98,
            hspace=0.02, wspace=0.05)

        # -- Plot time series

        plt.subplot(grid[0, 0])
        self.plot_timeseries("GHI", self.GHI, 1400)

        plt.subplot(grid[1, 0])
        self.plot_timeseries("DNI", self.DNI, 1400)

        plt.subplot(grid[2, 0])
        self.plot_timeseries("DIF", self.DIF, 1000)

        # -- Plot Heatmaps

        plt.subplot(grid[3, 0])
        self.plot_heatmap_timeseries("GHI", self.GHI, 700)

        plt.subplot(grid[4, 0])
        self.plot_heatmap_timeseries("DNI", self.DNI, 700)

        plt.subplot(grid[5, 0])
        self.plot_heatmap_timeseries("DIF", self.DIF, 700)

        # -- Plot ratios

        # XXX ? What does this do ?
        Smth = [1, 1]
        if Smth[0] < 1:
            xx = 0
            h1 = 1
        else:
            xx = np.linspace(-np.ceil(3 * Smth[0]), np.ceil(3 * Smth[0]), 2 * 3 * Smth[0] + 1)
            h1 = np.exp(-0.5 * (xx / Smth[0]) ** 2)
            h1 = h1 / sum(h1)
        if Smth[1] < 1:
            xx = 0
            h2 = 1
        else:
            xx = np.linspace(-3 * np.ceil(Smth[1]), np.ceil(3 * Smth[1]), 2 * 3 * Smth[1] + 1)
            h2 = np.exp(-0.5 * (xx / Smth[1]) ** 2)
            h2 = h2 / sum(h2)

        # DIF / GHI
        plt.subplot(grid[6, 0])
        self.plot_ratio_heatmap(
            ratios=self.DIF / self.GHI,
            filter=(self.DIF > 0) & (self.DNI > 0) & (self.GHI > 0),
            y_label='DIF/GHI (-)', title='Comparison of DIF and GHI for DNI<10W/m2. Should be close to 1.',
            ylimit=0.25, h1=h1, h2=h2)

        # GHI / estimated GHI
        plt.subplot(grid[7, 0])
        self.plot_ratio_heatmap(
            ratios=self.GHI / self.GHI_est,
            filter=(self.DNI > 0) & (self.GHI > 0) & (self.DNI < 5),
            y_label='GHI/(DNI*cSZA+DIF) (-)',
            title='Ratio of global to the sum of its components. Should be close to 1.',
            hlines=[(0.08, 0.8), (0.15, 1.0)], # (position relative to 1, linewidth)
            ylimit=0.25, h1=h1, h2=h2)

        # GHI / Clear sky
        if self.cams_df is not None:
            plt.subplot(grid[8, 0])
            self.plot_ratio_heatmap(
                ratios=self.GHI / self.cams_df.CLEAR_SKY_GHI,
                filter=(self.DNI > 0) & (self.GHI > 0),
                y_label='GHI/GHIcs (-)',
                title='Evaluation of McClear(*): Ratio of GHI to clear-sky GHI (GHIcs).',
                ylimit=0.75, bg_color=MC_CLEAR_COLOR,
                h1=h1, h2=h2)

            plt.annotate(
                '(*) not a plausibility control: the scatter points represent the joint effect of McClear and measurement errors.',
                (5, 2), xycoords='figure pixels',
                fontsize=6, fontstyle='italic', color=MC_CLEAR_COLOR)


        # =====================================================================
        # % % Part4: ERL& PPL tests
        # =====================================================================

        gs2 = GridSpec(4, 6)
        gs2.update(
            left=0.07, right=0.97,
            bottom=0.1, top=0.98,
            hspace=0.25, wspace=0.25)


        plt.subplot(gs2[0, 2])
        self.plot_bsrn_1c(self.GHI, "GHI", [[1.5, 1.2, 100], [1.2, 1.2, 50]])

        plt.subplot(gs2[1, 2])
        self.plot_bsrn_1c(self.DNI, "DNI", [[1, 0, 0], [0.95, 0.2, 10]])

        plt.subplot(gs2[2, 2])
        self.plot_bsrn_1c(self.DIF, "DIF", [[0.95, 1.2, 50],[0.75, 1.2, 30]])


        # BSRN 2C
        plt.subplot(gs2[0, 3])
        self.bsrn_2c()

        # SERI-Kn
        plt.subplot(gs2[1, 3])
        self.seri_kn()

        # SERI-K
        plt.subplot(gs2[2, 3])
        self.seri_k()

        # BSRN Closure
        plt.subplot(gs2[3, 2])
        self.bsrn_closure()

        # BRSN Closure ratio
        plt.subplot(gs2[3, 3])
        im = self.bsrn_closure_ratio()

        # Color legend
        cb_ax = fig.add_axes([0.38, 0.04, 0.28, 0.01])
        cbar = fig.colorbar(im, cax=cb_ax, orientation='horizontal', label='point density (-)')
        cbar.set_ticks([])

        # -- Third column

        # -- Text info
        self.print_info()

        # -- QC histograms
        info("QC: histograms of K, Kn & KT")

        gs3b = GridSpec(9, 9)
        gs3b.update(
            left=0.075,
            right=0.98,
            bottom=0.001,
            top=0.97,
            hspace=0.025,
            wspace=0.00)

        plt.subplot(gs3b[1:3, 6])
        self.histo_qc(self.GHI, self.flag_df.KT, 'GHI/TOA', y_label=True)

        plt.subplot(gs3b[1:3, 7])
        self.histo_qc(self.DNI, self.flag_df.Kn, 'DNI/TOANI')

        plt.subplot(gs3b[1:3, 8])
        self.histo_qc(self.DIF, self.flag_df.K, 'DIF/GHI',legend_pos='upper left')

        if self.cams_df is None :
            gs3 = GridSpec(7, 3)
            shadow_row = 3
        else:
            gs3 = GridSpec(9, 3)
            shadow_row = 5

        gs3.update(left=0.0, right=0.99, bottom=0.05, top=0.875, hspace=0.1, wspace=0.2)

        # -- Horizontality graph
        if self.cams_df is not None:

            info("Horizontality test")
            plt.subplot(gs3[3:5, 2])

            self.horizontality_graph()


        # -- Shadow analysis
        info("Shadow analysis (GHI)")

        plt.subplot(gs3[shadow_row:shadow_row+2, 2])
        self.shadow_analysis('GHI/TOA (-)', self.GHI, self.TOA, 0.85)

        plt.subplot(gs3[shadow_row + 2:shadow_row + 4, 2])
        self.shadow_analysis('DNI/TOANI (-)', self.DNI, self.TOANI, 0.65)



import pandas as pd

from libinsitu.common import GLOBAL_VAR, DIRECT_VAR, DIFFUSE_VAR, TEMP_VAR, HUMIDITY_VAR, PRESSURE_VAR, parseTimezone
from libinsitu.handlers.base_handler import map_cols, InSituHandler, ZERO_DEG_K

MAPPING = dict(
    GHI_corr_Avg=GLOBAL_VAR,
    DNI_corr_Avg=DIRECT_VAR,
    DHI_corr_Avg=DIFFUSE_VAR,
    Tair_Avg=TEMP_VAR,
    RH_Avg=HUMIDITY_VAR,
    BP_CS100_Avg=PRESSURE_VAR)

NA_VALUES= [-7999.0]

class ESMAPHandler(InSituHandler) :

    def _read_chunk(self, stream):
        df = pd.read_csv(stream, header=1, parse_dates=['TMSTAMP'], index_col=0, na_values=NA_VALUES)
        df = map_cols(df, MAPPING)

        # Update Time according to timezone
        df.index -= parseTimezone(self.properties["Station_Timezone"])

        df[HUMIDITY_VAR] = df[HUMIDITY_VAR] / 100  # percent -> 1
        df[PRESSURE_VAR] = df[PRESSURE_VAR] * 100 # Pressure hPa->Pa
        df[TEMP_VAR] = df[TEMP_VAR] + ZERO_DEG_K  # T2: °C -> K

        return df





#!/usr/bin/env python
# Temp file for merging input XLS file with CSV files
import argparse
import json
import re
import  os
from collections import defaultdict
from csv import DictReader
from glob import glob
from os.path import basename
import numpy as np
import netCDF4
import pytz
from timezonefinder import TimezoneFinder

from openpyxl import load_workbook
import csv
from unidecode import unidecode
from datetime import datetime
from urllib.request import urlretrieve

from libinsitu import VALID_COLS

OUT_DIR = "src/res/station-info/"

from libinsitu.log import error

FIRST_ROW = 2
FIRST_COL = 2

DATE_FORMAT="%Y-%m-%d"
ALTERNATE_FORMAT="%d/%m/%Y"

COL_SUBS = {
    "StationID" : "ID",
    "FullName" : "Name",
    "RawTimeZone" : "Timezone",
    "KoeppenGeigerClimate" : "Climate",
    "DNI" : "DNI_Col",
    "DHI" : "DHI_Col",
    "GHI" : "GHI_Col",
    "Type" : "QualityStandard"
}



GEOLOC = {
    "Country" : ["country"],
    "Region" : ["region", "state", "territory"],
    "Address": ["house_name", "house_number", "road"],
    "City" : ["municipality", "city", "town", "village"]
}

LOCKED_COLS = ["StartDate", "EndDate", "TimeResolution"]

# For ID generation
FULLNAME_REPL = {
    "ST." : "",
    "SCHWABISCH" : "SCH"
}

def idTransformer(value) :
    return value.upper()

TRANSFORMERS = {
    "ID" : idTransformer,
}

colCounter = defaultdict(lambda : 0)


def cleanColName(col) :
    col = col.replace("Station_", "")
    if col in COL_SUBS :
        col = COL_SUBS[col]
    col = col[0].upper() + col[1:]
    return col

COMMA_NUMBER = r"^[0-9,]*$"


def str2val(val) :
    try:
        date = datetime.strptime(ALTERNATE_FORMAT, val)
        return date.strftime(DATE_FORMAT)
    except:
        pass

    if re.match(COMMA_NUMBER, val):
        val = val.replace(",", ".")
    try:
        return int(val)
    except:
        try:
            fval = float(val)
            if fval.is_integer():
                return int(fval)
            else:
                return fval
        except:
            return val

def get_value(cell):

    val = cell.value

    if val is None :
        return ""

    if cell.is_date :
        return val.strftime(DATE_FORMAT)

    elif isinstance(val, str) :
        return str2val(val)
    else:
        return cell.value

def read_sheet_raw(sheet) :
    """Read raw sheet into a list of dict"""

    cols_idx = dict((sheet.cell(FIRST_ROW, col_idx).value, col_idx) for col_idx in range(FIRST_COL, sheet.max_column+1))
    cols_idx = dict((cleanColName(col), idx) for col, idx in cols_idx.items() if col is not None)

    # Filter and sort columns
    sorted_cols = sorted(list(col for col in cols_idx.keys() if col in VALID_COLS), key=lambda col : VALID_COLS.index(col))

    # Filter valid co land sort by column order
    sorted_cols_idx = dict((col, cols_idx[col]) for col in sorted_cols)
    res = []

    for row_idx in range(FIRST_ROW+1, sheet.max_row+1) :
        row = dict((col, get_value(sheet.cell(row_idx, col_idx))) for col, col_idx in sorted_cols_idx.items())

        # Skip empty rows
        if all(val == "" for key, val in row.items()) :
            continue

        res.append(row)
    return res


def generate_id(row) :

    if "ID" not in row or row["ID"] == "" :
        fullname = row["Name"]
        fullname = unidecode(fullname).upper()
        for key, repl in FULLNAME_REPL.items():
            fullname = fullname.replace(key, repl).strip()
        fullname = fullname.replace(' ', "")

        id = fullname[0:4]
        print("Generate ID: '%s' -> '%s'" % (fullname, id))

        row["ID"] = id

def count_cols(rows) :
    for row in rows :
        for key, val in row.items() :
            if val is not None and val != "" :
                colCounter[key] +=1
            else:
                colCounter[key] += 0

def read_network(network, sheet) :

    rows = read_sheet_raw(sheet)

    # Transform values if needed
    for row_idx, row in enumerate(rows) :
        for key, transformer in TRANSFORMERS.items() :
            try:
                if key in row :
                    row[key] = transformer(row[key])

                generate_id(row)

                if row["ID"] == "" :
                    raise Exception("Empty ID")

            except Exception as e :
                error("Happened in %s: line %d, col:%s" % (network, row_idx, key))
                raise e

    # Check IDS are unique
    ids = set()
    for row in rows :
        id = row["ID"]
        if id in ids :
            raise Exception("Duplicate ID for network %s : %s" % (network, id))
        ids.add(id)

    count_cols(rows)

    return rows

def read_networks(workbook):
    dir(workbook)
    networks = workbook.sheetnames
    res = dict()

    for network in networks:
        network = network.strip()
        if network == "OverviewNetworks":
            continue

        res[network] = read_network(network, workbook[network])

    return res

def load_csv(path) :
    if not os.path.exists(path) :
        return []
    with open(path, 'r') as f :
        reader = DictReader(f)
        return list(reader)

def filter_redorder_cols(rows) :
    cols = list(rows[0].keys())
    filtered_cols = list(col for col in cols if col in VALID_COLS)

    if len(filtered_cols) < len(cols) :
        removed = set(cols) - set(filtered_cols)
        print("Removed columns: %s" % str(removed))

    sorted_cols = sorted(filtered_cols, key=lambda col : VALID_COLS.index(col))

    res = []
    for row in rows :
        res.append({col: row[col] for col in sorted_cols})
    return res

def save_csv(path, rows) :
    with open(path, "w") as f:
        writer = csv.DictWriter(f, fieldnames=list(rows[0].keys()), delimiter=",", lineterminator=os.linesep)
        writer.writeheader()
        for row in rows:
            writer.writerow(row)

def merge(initial_rows, updated_rows) :

    initial_by_id = dict((row["ID"], row) for row in initial_rows)

    for row in updated_rows :
        id = row["ID"]
        if not id in initial_by_id :
            initial_by_id[id] = row
        else:
            initial_row = initial_by_id[id]

            for col, val in row.items() :
                if not col in initial_row :
                    initial_row[col] = val
                else:
                    initial_val = initial_row[col]
                    if str2val(initial_val) != val and col not in LOCKED_COLS :
                        print("Updating %s#%s : %s => %s " % (id, col, initial_val, val))
                        initial_row[col] = val

    print("Cols after merge", initial_rows[0].keys())

    return list(initial_by_id.values())



def save(networks, out_folder) :
    """Update existing CSV or create one"""
    for network, rows in networks.items() :

        print("Processing network %s" % network)

        path = os.path.join(out_folder, network + ".csv")

        initial = load_csv(path)

        initial = filter_redorder_cols(initial)

        merged = merge(initial, rows)

        save_csv(path, merged)

def get_loc(network, id, lat, lon) :
    filename = os.path.join(CACHE_FOLDER, "%s-%s.js" % (network, id))
    if not os.path.exists(filename) :
        url ="https://nominatim.openstreetmap.org/reverse?lat=%f&lon=%f&format=json&accept-language=en" % (lat, lon)
        urlretrieve(url, filename)

    with open(filename, "r") as f :
        return json.load(f)

def enrich_address(network, row, lat, lon) :

    loc = get_loc(network, row["ID"], lat, lon)

    if "error" in loc:
        print("Error for : %s/%s : %s" % (network, id, loc["error"]))
        return

    address = loc["address"]

    for col, keys in GEOLOC.items():
        val = ", ".join(address[key] for key in keys if key in address)
        #val = unidecode(val)
        row[col] = val
        #print("%s#%s : %s" % (id, col, val))

def enrich_climate(nc, row, lat, lon) :
    climate = get_KG_ClimZone(nc, lat, lon)
    row["Climate"] = climate

def enrich_timezone(tzFinder, row, lat, lon) :
    tzname = tzFinder.timezone_at(lng=lon, lat=lat)
    tz = pytz.timezone(tzname)
    now = datetime.now(tz)

    offset_min = now.utcoffset().total_seconds() / 60

    sign="+"
    if offset_min < 0 :
        sign = "-"
        offset_min = - offset_min

    row["Timezone"] = "UTC%s%02d:%02d" % (sign, offset_min / 60, offset_min % 60)

def enrich_coords(network, rows, nc_climate, ids=None) :

    tzFinder = TimezoneFinder()

    for row in rows:

        if ids is not None and row["ID"] not in ids :
            continue

        lat = str2val(row["Latitude"])
        lon = str2val(row["Longitude"])

        enrich_address(network, row, lat, lon)
        enrich_climate(nc_climate, row, lat, lon)
        enrich_timezone(tzFinder, row, lat, lon)

    return rows


def main_xls(xls_file, out_folder) :

    wb = load_workbook(xls_file)
    networks = read_networks(wb)

    save(networks, out_folder)

    # Show empty cols
    for col, count in colCounter.items():
        print("Count '%s' : %d" % (col, count))


def get_KG_ClimZone(ds, lat, lon):
    vlat = ds['lat'][:]
    dlat = np.abs(vlat - lat)
    ilat = np.where(dlat == min(dlat))[0][0]

    if (lon > 180):
        lon += -360
    vlon = ds['lon'][:]
    dlon = np.abs(vlon - lon)
    ilon = np.where(dlon == min(dlon))[0][0]

    ID = np.minimum(30, ds['Band1'][ilat, ilon].data.astype(int))
    if (ID == 0):
        return
    else:
        return ds.getncattr(str(ID))


def main_coords(out_folder, climate_file, network=None, ids=None) :


    nc_climate = netCDF4.Dataset(climate_file) if climate_file else None

    for file in glob(os.path.join(out_folder, "*.csv")) :

        base = basename(file)

        if network != None and base != network + ".csv" :
            continue

        print("Processing %s" % base)

        rows = load_csv(file)

        network = basename(file)
        network = network.replace(".csv", "")

        if nc_climate :
            rows = enrich_coords(network, rows, nc_climate, ids)

        rows = filter_redorder_cols(rows)

        save_csv(file, rows)


def main() :

    global CACHE_FOLDER

    parser = argparse.ArgumentParser(description='Enrich CSV files of stations')
    subparsers = parser.add_subparsers(help="commands", dest="command", required=True)

    parser.add_argument('--out-folder', "-o", metavar='<output-folder>', type=str, help='Output folder', default=OUT_DIR)

    xls_parser = subparsers.add_parser('xls', help="Enrich data from single XLS file")
    xls_parser.add_argument("input_file", metavar="<input.xls>", type=str)

    coords_parser = subparsers.add_parser('coords', help="Enrich data from coordinates")
    coords_parser.add_argument("--cache", "-c", metavar="<tmp_dir>", type=str, help="Cache folder", default="/tmp")
    coords_parser.add_argument("--climate", "-cli", metavar="<climate.nc>", type=str, help="NetCDF climate file")
    coords_parser.add_argument("--network", "-net", metavar="<network>", type=str, help="Only process a single network")
    coords_parser.add_argument("--stations-ids", "-ids", metavar="<id1>,<ids2>,...", type=str, help="Only process given stations")
    args = parser.parse_args()

    if args.command == "xls" :
        main_xls(args.input_file, args.out_folder)
    elif args.command == "coords" :

        CACHE_FOLDER = args.cache

        ids = None if args.stations_ids is None else args.stations_ids.split(",")

        main_coords(args.out_folder, args.climate, args.network, ids)

    else:
        raise Exception("Unknown command" + args.command)

if __name__ == '__main__':
    main()


import pandas as pd

from libinsitu.common import GLOBAL_VAR, DIRECT_VAR, DIFFUSE_VAR, parseTimezone
from libinsitu.handlers.base_handler import InSituHandler


class ABOMHandler(InSituHandler) :

    def _read_chunk(self, stream) :

        def date_parser(years, months, days, hours, minutes):
            strs = years + "/" + months + "/" + days + " " + hours + ":" + minutes
            return pd.to_datetime(strs)

        data = pd.read_csv(
            stream, skipinitialspace=True, index_col="datetime",
            parse_dates=dict(datetime=[2, 3, 4, 5, 6]), date_parser=date_parser)

        ghi_col = data.columns[2]
        dir_col = data.columns[7]
        dif_col = data.columns[12]

        mapping = {
            ghi_col:GLOBAL_VAR,
            dir_col:DIRECT_VAR,
            dif_col:DIFFUSE_VAR}

        data = data[list(mapping.keys())]
        data = data.rename(columns=mapping)

        # Apply timezone
        data.index -= parseTimezone(self.properties["Station_Timezone"])

        return data

    def data_vars(self):
        """ @override """
        return [GLOBAL_VAR, DIFFUSE_VAR, DIRECT_VAR]
